"""
Helpers for plotting of maps etc
--------------------------------

.. note::

    BETA stage, code might undergo revisions

"""
import numpy as np
import matplotlib.pyplot as plt
from cartopy import crs
from cartopy.mpl.ticker import LongitudeFormatter, LatitudeFormatter

from geonum.helpers import order_of_magnitude


def init_figure_with_geoaxes(projection=None,
                             xlim=(-180,180),
                             ylim=(-90, 90)):
    """
    Initiate a matplotlib figure containing 1 cartopy GeoAxes instance

    Can be used to plot maps.

    Parameters
    ----------
    projection : cartopy.crs.CRS, optional
        Projection used, if None, then :class:`cartopy.crs.PlateCarree` is
        used. Defaults to None.
    xlim : tuple, optional
        longitude range of map, defaults to whole globe (-180,180).
    ylim : tuple, optional
        latitude range of map, defaults to whole globe (-90,90).

    Returns
    -------
    cartopy.mpl.geoaxes.GeoAxes

    """
    if projection is None:
        projection = crs.PlateCarree()
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1, projection=projection)
    ax.set_xlim(xlim)
    ax.set_ylim(ylim)
    return ax

def _get_tick_formatter(value):
    """
    Get tick formatting string for map plots

    Parameters
    ----------
    value : float
        value used to estimate formatter (e.g. one of the longitude
        coordinates).

    Returns
    -------
    str
        formatting string

    """
    digits = 2 - order_of_magnitude(value)
    digits = 0 if digits < 0 else digits
    return f'.{digits:d}f'

def _calculate_map_ticks(ax):
    """
    Calculate longitude (x) and latitude (y) ticks for map plots.

    Parameters
    ----------
    ax : cartopy.mpl.geoaxes.GeoAxes
        map axes instances.

    Returns
    -------
    tuple
        2-element tuple containing x and y ticks
    """
    lonleft, lonright = ax.get_xlim()

    num_lonticks = 7 if lonleft == -lonright else 6
    xticks = np.linspace(lonleft, lonright, num_lonticks)

    latleft, latright = ax.get_ylim()
    num_latticks = 7 if latleft == - latright else 6
    yticks = np.linspace(latleft, latright, num_latticks)
    return (xticks, yticks)

def set_map_ticks(ax, xticks=None, yticks=None, tick_format=None):
    """Set or update ticks in instance of GeoAxes object (cartopy)

    Note
    ----
    The input GeoAxes must be setup with :class:`cartopy.crs.PlateCarree`
    projection.

    Parameters
    ----------
    ax : cartopy.GeoAxes
        map axes instance (needs to be set up with PlateCarree projection).
    xticks : iterable, optional
        ticks of x-axis (longitudes). If None, it is retrieved automatically
        based on xlim of input axes. Defaults to None.
    yticks : iterable, optional
        ticks of y-axis (latitudes). If None (or if xticks is None),
        it is retrieved automatically based on ylim of input axes. Defaults
        to None.
    tick_formatter : str, optional
        string formatter for axes labels in the plot.

    Returns
    -------
    cartopy.GeoAxes
        modified axes instance
    """
    if not isinstance(ax.projection, crs.PlateCarree):
        raise AttributeError('Map ticks can only be added to GeoAxes with '
                             'PlateCarree projection.')

    if xticks is None:
        xticks, yticks = _calculate_map_ticks(ax)

    if tick_format is None:
        val = np.mean(xticks)
        if val == 0:
            val = xticks[next((i for i, x in enumerate(xticks) if x), None)]
        if val != 0:
            tick_format = _get_tick_formatter(val)
        else:
            tick_format = '.1f'

    ax.set_xticks(xticks, crs=crs.PlateCarree())
    ax.set_yticks(yticks, crs=crs.PlateCarree())

    lon_formatter = LongitudeFormatter(number_format=tick_format,
                                       degree_symbol='°',
                                       dateline_direction_label=True)
    lat_formatter = LatitudeFormatter(number_format=tick_format,
                                      degree_symbol='°')

    ax.xaxis.set_major_formatter(lon_formatter)
    ax.yaxis.set_major_formatter(lat_formatter)

    return ax

def _infer_text_location(ax,p):
    info = dict(ha='center')
    left,right = ax.get_xlim()
    dx = right - left
    bottom,top = ax.get_ylim()
    dy = top - bottom
    lat, lon = p.latitude, p.longitude
    distleft = abs(left - lon) / dx
    if distleft < 0.3:
        info['ha'] = 'left'
    elif distleft > 0.7:
        info['ha'] = 'right'

    yoffs = dy*0.08 # put text 5% of range above or below the points latitude
    distbottom = abs(bottom - lat) / dx
    if distbottom > 0.6: # point is in upper 40% of map, put text below
        yoffs *= -1
    info['xytext'] = (lon, lat+yoffs)
    return info

def plot_topo_contourf(ax, topo, oceans_separate=True, levels=50, cmap=None):
    if cmap is None:
        cmap = 'Oranges'

    # Create meshgrid for domain for contour plot
    X,Y = topo.init_mesh()

    # Separate land areas from sea areas for plotting below
    if oceans_separate:
        seamask = np.ma.masked_less_equal(topo.data, 0)
        data = np.ma.MaskedArray(topo.data, mask=seamask.mask)
        if np.any(seamask):
            seadata = np.ma.MaskedArray(topo.data, mask=~seamask.mask)
            # Plot sea in a light blue
            ax.contourf(X, Y, seadata, 50,
                        colors='#e6f7ff')
    else:
        data = topo.data

    # Plot topographic land data
    pdata = ax.contourf(X, Y, data, levels=levels,
                cmap=cmap)

    return ax, pdata

def plot_geopoint_into_map(ax, p, annotate=True, annot_kwargs=None, **kwargs):

    if annot_kwargs is None:
        annot_kwargs = {}

    ax.scatter(x=[p.longitude],y=[p.latitude], **kwargs)

    if annotate:
        annot_loc = _infer_text_location(ax,p)
        try:
            c = kwargs['color']
        except KeyError:
            c = 'k'
        annot = dict(
            text = p.name,
            xy = (p.longitude,
                  p.latitude),  # location of observatory in plot
            arrowprops = dict(color=c, lw=1, arrowstyle='->', shrinkB=4),
            size=7, color=c, fontweight='bold', **annot_loc
        )
        annot.update(annot_kwargs)
        ax.annotate(**annot)
    return ax

def plot_geovector3d_into_map(ax, vec, **kwargs):
    if vec.anchor is None:
        raise AttributeError(
            f'input vector {vec} does not have an anchor point assigned. '
            f'Please set anchor location using vec.add_anchor')

    start = vec.anchor
    end = start + vec
    ax.plot([start.longitude, end.longitude],
            [start.latitude, end.latitude], **kwargs)
    return ax


# -*- coding: utf-8 -*-
#
# Geonum is a Python library for geographical calculations in 3D
# Copyright (C) 2017 Jonas Gliss (jonasgliss@gmail.com)
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License a
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from geonum import CV2_AVAILABLE
from LatLon23 import LatLon

class TopoData(object):
    """Data class for topography data

    This object represents topographic data in a certain latitude and longitude
    range, specified by the corner coordinates of the range
    (:attr:`lon0, lat0, lon1, lat1`). It may be used for 2D and 3D plotting of
    the topographic data (cf. :func:`plot_2d`, :func:`plot_3d`) or for further
    analysis such as coordinate based altitude retrievals
    (:func:`get_altitude`), change of the grid resolution (cf.
    :func:`increase_grid_resolution`) or for customised analysis by directly
    using the numpy data array containing the topographic altitudes
    (:attr:`data`), together with the corresponding dimension coordinates
    which can be accessed via attributes :attr:`latitude`and :attr:`longitude`.

    To create an instance of this class, you may use the class
    :class:`TopoDataAccess`.

    Note
    ----
    This object does not provide topography data access. For data access,
    please use :class:`TopoDataAccess` which returns :class:`TopoData`
    objects. This object is intended for storage and post processing of
    topographic data

    Parameters
    ----------
    lats : ndarray
        numpy array with latitude coordinates of the topographic dataset.
        Accessible via :attr:`latitude`.
    lons : ndarray
        numpy array with longitude coordinates of the topographic dataset.
        Accessible via :attr:`longitude`.
    data : ndarray
        2D numpy array containing elevation values, first index denotes
        latitude dimension, 2nd index denotes longitude dimension.
    data_id : str
        ID of this data set.
    repl_nan_minval : bool
        if True, then coordinates containing NaN values are replaced with the
        minimum altitude in `data`.
    """
    def __init__(self, lats, lons, data, data_id=None, repl_nan_minval=False):
        if data_id is None:
            data_id = 'undefined'
        self.data_id = data_id

        lats = np.asarray(lats)
        lons = np.asarray(lons)
        data = np.asarray(data)
        if not data.shape == (len(lats), len(lons)):
            raise ValueError(
                'shape mismatch between input lats and lons and data...'
                )

        self.lats = lats
        self.lons = lons
        self.data = data


        if repl_nan_minval:
            self.replace_nans()

    def __str__(self):
        return (
            f'geonum.TopoData (data_id: {self.data_id})\n'
            f'Dimensions: {self.dims}, shape: {self.shape}\n'
            f'latitude range: {self.lat0} - {self.lat1}\n'
            f'longitude range: {self.lon0} - {self.lon1}\n'
            f'data: {type(self.data)}\n{self.data}'
            )

    def __repr__(self):
        return (f'geonum.TopoData (data_id: {self.data_id}, shape: {self.shape})\n')

    @property
    def latitude(self):
        """Wrapper for :attr:`lats`"""
        return self.lats

    @property
    def longitude(self):
        """Wrapper for :attr:`lons`"""
        return self.lons

    @property
    def dims(self):
        """
        list: List of dimension names
        """
        return ['latitude', 'longitude']

    def replace_nans(self, fillval=None):
        """Replace NaNs in topographic data with a fill value

        Parameters
        ----------
        fillval : float
            value that is assigned to all coordinates that include NaNs
        """
        if fillval is None:
            fillval = np.nanmin(self.data)
        self.data[np.isnan(self.data)] = fillval

    @property
    def lon0(self):
        """Returns first longitude (i.e. ``self.lons[0]``)"""
        return self.lons[0]

    @property
    def lat0(self):
        """Returns first latitude (i.e. ``self.lats[0]``)"""
        return self.lats[0]

    @property
    def lon1(self):
        """Returns last longitude (i.e. ``self.lons[-1]``)"""
        return self.lons[-1]

    @property
    def lat1(self):
        """Returns last latitude (i.e. ``self.lats[-1]``)"""
        return self.lats[-1]

    @property
    def max(self):
        """Returns maximum altitude of topo data"""
        return np.nanmax(self.data)

    @property
    def min(self):
        """Returns minimum altitude of topo data"""
        return np.nanmin(self.data)

    @property
    def shape(self):
        """Shape of 2D numpy array that contains topography"""
        return self.data.shape

    @property
    def alt_range(self):
        """Returns covered altitude range"""
        return self.max - self.min

    @property
    def resolution(self):
        """Returns tuple (lat, lon) of the current grid resolution in km

        Note
        ----
        The resolution is determined at the center of this grid
        """
        if not len(self.lons) > 1 or not len(self.lats) > 1:
            raise ValueError('Need at least 2x2 sized topofield')
        x_lon, x_lat = int(len(self.lons) / 2), int(len(self.lats) / 2)
        p0 = LatLon(self.lats[x_lat], self.lons[x_lon])
        r_lon = (p0 - LatLon(self.lats[x_lat], self.lons[x_lon-1])).magnitude
        r_lat = (p0 - LatLon(self.lats[x_lat-1], self.lons[x_lon])).magnitude
        return (r_lat, r_lon)

    def mean(self):
        """Mean elevation (ignores NaNs in data)"""
        return np.nanmean(self.data)

    def std(self):
        """Standard deviation of topographic dataset"""
        return np.nanstd(self.data)

    def increase_grid_resolution(self, res=0.2, polyorder=2): # pragma: no cover
        """Gaussian pyramide based upscaling of topographic grid

        This function checks the current topographic resolution in the center
        of the grid. Based on this, an upsacling factor is determined and the
        :func:`cv2.pyrUp` is used to increase the resolution using
        interpolation. Note, that this does not increase the actual resolution
        of the topographic data grid.

        Note
        ----
        Requires that opencv library is installed.

        Parameters
        ----------
        res : int or float, optional
            desired grid resolution in km (default: 0.2)
        polyorder : int, optional
            order of polynomial used for interpolation (default: 2)

        Returns
        -------
        TopoData
            new object with desired grid resolution

        Raises
        ------
        ImportError
            if opencv is not installed.
        """
        if not CV2_AVAILABLE:
            raise ImportError('Feature disabled: Require opencv')
        from cv2 import pyrUp
        lons = self.lons
        lats = self.lats
        vals = self.data
        if not all(np.mod(x, 2) == 0 for x in [len(lons), len(lats)]):
            print ("Fatal error, odd array size detected, no upscaling possible."
                " Return current data dict")
            return False

        c_lon = len(lons) / 2 - 1 #center longitude index
        c_lat = len(lats) / 2 - 1 #center latitude index
        p1 = LatLon(lats[c_lat], lons[c_lon])
        p2 = LatLon(lats[c_lat + 1], lons[c_lon + 1])
        dist = (p2 - p1).magnitude #distance between 2 points on grid
        res_fac = dist / res #factor for increasing the spatial resolution
        if res_fac <= 1:
            print("No interpolation of topodata necessary: topo raw data "
                  "already has desired resolution...")
            return self
        fac = int(np.ceil(np.log2(res_fac))) #the corresponding up-factor for the scale space
        print(("Increasing spatial topography resolution by factor %s" %(2**fac)))
        for k in range(fac):
            vals = pyrUp(vals)
        p_lons = np.poly1d(np.polyfit(np.arange(len(lons)), lons, polyorder))
        p_lats = np.poly1d(np.polyfit(np.arange(len(lats)), lats, polyorder))
        lons_new = p_lons(np.linspace(0, len(lons) - 1, vals.shape[1]))
        lats_new = p_lats(np.linspace(0, len(lats) - 1, vals.shape[0]))
        return TopoData(lats_new, lons_new, vals, self.data_id + "_interp")

    @property
    def delta_lon(self):
        """Longitude range of data (in decimal degrees)"""
        return self.lons[-1] - self.lons[0]

    @property
    def delta_lat(self):
        """Latitude range of data (in decimal degrees)"""
        return self.lats[-1] - self.lats[0]

    @property
    def center_coordinates(self):
        """Tuple (lat, lon) with center coordinates of data"""
        return (self.lats[0] + self.delta_lat / 2.,
                self.lons[0] + self.delta_lon / 2.)

    def init_mesh(self):
        """
        Init X,Y meshgrid from latitudes and longitudes for map plots

        Returns
        -------
        list
            meshgrid

        """
        return np.meshgrid(self.longitude, self.latitude)

    def plot(self, plot3d=True, draw_coastlines=False,
             draw_mapscale=False, **kwargs): # pragma: no cover
        from geonum.mapping import Map
        if not "projection" in kwargs:
            kwargs["projection"] = "lcc"
        if not "llcrnrlon" in kwargs:
            kwargs["llcrnrlat"] = self.lat0
            kwargs["llcrnrlon"] = self.lon0
            kwargs["urcrnrlat"] = self.lat1
            kwargs["urcrnrlon"] = self.lon1

            kwargs["lat_0"] = self.lat0 + (self.lat1-self.lat0)/2
            kwargs["lon_0"] = self.lon0 + (self.lon1-self.lon0)/2

        m = Map(**kwargs)
        m.set_topo_data(self)

        if plot3d:
            if draw_coastlines:
                raise NotImplementedError('3D plotting and coastline drawing is '
                                          'not supported yet')
            elif draw_mapscale:
                raise NotImplementedError('3D plotting and mapscale drawing is '
                                          'not supported yet')
            m.draw_topo_3d()
        else:
            if draw_coastlines:
                m.draw_coastlines()
            m.draw_coordinates()
            if draw_mapscale:
                m.draw_mapscale_auto()
        return m

    def plot_2d(self, ax=None): # pragma: no cover
        """Plot 2D basemap of topodata"""
        from geonum.mapping import Map
        latc, lonc = self.center_coordinates
        m = Map(self.lon0, self.lat0, self.lon1, self.lat1,
                projection="merc", lat_0=latc, lon_0=lonc, ax=ax)
        m.set_topo_data(self)
        m.draw_topo(insert_colorbar=True)
        m.draw_topo_contour()
        m.draw_coordinates()
        m.drawcoastlines()
        return m

    def plot_3d(self, ax=None): # pragma: no cover
        """Creates 3D surface plot of data

        Parameters
        ----------
        ax
            instance of matplotlib Axes3D object

        Returns
        -------
        geonum.mapping.Map
            plotted map object.

        """
        from geonum.mapping import Map
        latc, lonc = self.center_coordinates
        m = Map(self.lon0, self.lat0, self.lon1, self.lat1,
                projection="merc", lat_0=latc, lon_0=lonc, ax=ax)
        m.set_topo_data(self)
        m.draw_topo_3d()
        return m

    def includes_coordinate(self, lat=None, lon=None):
        """Checks if input coordinates are covered by this dataset"""
        if not (self.lat0 <= lat <= self.lat1 and
                self.lon0 <= lon <= self.lon1):
            print("Input coordinates out of topo data range...")
            return False
        return True

    def closest_index(self, lat, lon):
        """Finds the closest index to input coordinate

        Parameters
        ----------
        lat : float
            latitude coordinate in decimal degrees
        lon : float
            longitude coordinate in decimal degrees

        Returns
        -------
        tuple
            2-element tuple containing the closest index of lat and lon
            arrays to input index

        Raises
        ------
        ValueError
            if input coordinate is not included in this dataset
        """
        if not self.includes_coordinate(lat, lon):
            raise ValueError("Input values out of range...")

        idx_lat = np.argmin(abs(self.lats - lat))
        idx_lon = np.argmin(abs(self.lons - lon))
        return (idx_lat, idx_lon)

    def get_altitude(self, lat, lon):
        """Get altitude value at input coordinate

        Parameters
        ----------
        lat : int or float
            latitude of coordinate
        lon : int or float
            longitude of coordinate

        Returns
        -------
        float
            altitude at input coordinate
        """
        idx_lat, idx_lon = self.closest_index(lat, lon)
        return self.data[idx_lat, idx_lon]

    def __call__(self, lat, lon):
        """Get altitude value at input position"""
        try:
            return self.get_altitude(lat, lon)
        except ValueError as e:
            raise e


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Mar  6 21:15:34 2021

@author: jonasg
"""
import pytest
import numpy as np
import numpy.testing as npt

from geonum import topodata as td
from geonum.conftest import does_not_raise_exception
arr = np.array([[647., 647.], [651., 651.]])
lats = np.array([48., 48.00083333])
lons = np.array([11., 11.00083333])

@pytest.fixture(scope='module')
def topodata():
    return td.TopoData(
        np.array([48., 48.00083333]),
        np.array([11., 11.00083333]),
        np.array([[647., 647.], [651., 651.]]),None,False)

@pytest.mark.parametrize('lats,lons,data,data_id,repl_nan_minval,raises', [
        (np.array([48., 48.00083333]), np.array([11., 11.00083333]),
         np.array([[647., 647.], [651., 651.]]),None,False,
         does_not_raise_exception()),
        (np.array([48., 48.00083333]), np.array([11., 11.00083333]),
         np.array([[647., 647.], [651., np.nan]]),None,False,
         does_not_raise_exception()),
         (np.array([48., 48.00083333]), np.array([11., 11.00083333]),
         np.array([[647., 647.], [651., np.nan]]),None,True,
         does_not_raise_exception()),
        (np.array([48., 48.00083333]), np.array([11., 11.00083333]),
         np.array([[647., 647.], [651., np.nan]]),'great_data',True,
         does_not_raise_exception()),
        ([1,2,3],[1,2],[[1,2],[1,2]],None,False,pytest.raises(ValueError)),
        ([1,2],[1,2,3],[[1,2,3],[1,2,3]],None,False,does_not_raise_exception()),
        ([1,2,3],[1,2],[[1,2,3],[1,2,3]],None,False,pytest.raises(ValueError))
        ])
def test__init__(lats, lons, data, data_id, repl_nan_minval,
                          raises):
    with raises:
        topo = td.TopoData(lats, lons, data, data_id, repl_nan_minval)
        npt.assert_array_equal(topo.lats, lats)
        npt.assert_array_equal(topo.lons, lons)

        if data_id is None:
            data_id='undefined'
        assert topo.data_id == data_id
        if repl_nan_minval:
            data[np.isnan(data)] = np.nanmin(arr)
        npt.assert_array_equal(topo.data, data)

def test_latitude(topodata):
    assert topodata.latitude is topodata.lats

def test_longitude(topodata):
    assert topodata.longitude is topodata.lons

def test_dims(topodata):
    assert topodata.dims == ['latitude', 'longitude']

def test___str__(topodata):
    assert str(topodata).startswith('geonum.TopoData')

def test___repr__(topodata):
    assert repr(topodata).startswith('geonum.TopoData')

def test_replace_nans():
    topo = td.TopoData(
        np.array([48., 48.00083333]),
        np.array([11., 11.00083333]),
        np.array([[647., 647.], [np.nan, 651.]]),None,False)
    topo.replace_nans(1000.)
    assert not np.any(np.isnan(topo.data))

def test_alt_range(topodata):
    assert topodata.alt_range == 4

def test_resolution_ERR():
    d = td.TopoData([1],[1],[[1]])
    with pytest.raises(ValueError):
        d.resolution
def test_resolution(topodata):
    np.testing.assert_allclose(topodata.resolution, (0.09,0.06), atol=0.01)

def test_mean(topodata):
    assert topodata.mean() == 649

def test_std(topodata):
    assert topodata.std() == 2
def test_delta_lon(topodata):
    np.testing.assert_allclose(topodata.delta_lon, 0.000833, atol=1e-5)

def test_delta_lat(topodata):
    np.testing.assert_allclose(topodata.delta_lat, 0.000833, atol=1e-5)

def test_center_coordinates(topodata):
    np.testing.assert_allclose(
        topodata.center_coordinates,
        (48+0.000833/2, 11+0.000833/2),
        atol=1e-5)

def test_init_mesh(topodata):
    X,Y = topodata.init_mesh()
    assert X.shape == Y.shape == (2,2)

def test_closest_index_OUTOFBOUNDS(topodata):
    with pytest.raises(ValueError):
        topodata.closest_index(0,0)

def test_closest_index(topodata):
    assert topodata.closest_index(48,11) == (0,0)

def test___call__ERR(topodata):
    with pytest.raises(ValueError):
        topodata(0,0)

def test___call__(topodata):
    assert topodata(48,11) == 647

if __name__ == "__main__":
    import sys
    pytest.main(sys.argv)


# Geonum is a Python library for geographical calculations in 3D
# Copyright (C) 2017 Jonas Gliss (jonasgliss@gmail.com)
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License a
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import os

import numpy as np

from geonum import TOPO_INFO_FILE, LOCAL_TOPO_DIR

def order_of_magnitude(num):
    """
    Get exponent precision of input number

    E.g. 1 if input number is 14.2, 0 if input number is 3 or -1 if input
    number is 0.1.

    Note
    ----
    The order of magnitude here is calculated using floor(log10(num)), thus,
    num=9, would result in a value of 0, in contrary to other definitions (
    e.g. https://en.wikipedia.org/wiki/Order_of_magnitude).

    Parameters
    ----------
    num : float
        input number

    Returns
    -------
    int
        order of magnitude of input number

    """
    return int(np.floor(np.log10(np.abs(num))))


def all_topodata_search_dirs():
    """Returns a list of all directories that are searched for topography data

    Returns
    -------
    list
        list containing all search directories (absolute paths)
    """
    if LOCAL_TOPO_DIR is None:
        raise FileNotFoundError('geonum local topo directory not accessible')
    paths = []
    with open(TOPO_INFO_FILE, "r") as f:
        for line in f:
            p = line.strip()
            if os.path.exists(p) and not p in paths:
                paths.append(os.path.normpath(p))
    return paths


def check_and_add_topodir(local_dir):
    """Check if input directory is registered and if not, register it

    Parameters
    ----------
    local_dir : str
        directory that is supposed to be checked

    """
    fp = os.path.normpath(local_dir)
    if not os.path.exists(fp):
        raise ValueError(f'{local_dir} does not exist...')
    elif not fp in all_topodata_search_dirs():
        with open(TOPO_INFO_FILE, "a") as f:
            f.write(f'{fp}\n')
            print(f'Adding {fp} to file ~/.geonum/LOCAL_TOPO_DATA')


def isnum(val):
    """Checks if input is number (int or float) or and not nan

    Parameters
    ----------
    val
        input object to be checked

    Returns
    -------
    bool
        whether or not input object `val` is of numerical type
    """
    if isinstance(val, (int, float)) and not np.isnan(val):
        return True
    return False


# ToDo: review and move to plot_helpers.py
def rotate_xtick_labels(ax, deg=30, ha="right"):
    """Rotate xtick labels in matplotlib axes object"""

    lbls = ax.get_xticklabels()
    lbls = [lbl.get_text() for lbl in lbls]
    ax.set_xticklabels(lbls, rotation=deg, ha=ha)
    return ax


def haversine_formula(lon0, lat0, lon1, lat1, radius=None):
    """Haversine formula to compute distances on a sphere

    Approximate horizontal distance between 2 points assuming a spherical
    earth.

    Parameters
    ----------
    lon0 : float
        longitude of first point in decimal degrees
    lat0 : float
        latitude of first point in decimal degrees
    lon1 : float
        longitude of second point in decimal degrees
    lat1 : float
        latitude of second point in decimal degrees
    radius : float
        average earth radius in km, defaults to 6371 km

    Returns
    -------
    float
        distance of both points in km
    """
    if radius is None:
        radius = 6371

    hav = lambda d_theta: np.sin(d_theta / 2.0) ** 2

    d_lon = np.radians(lon1 - lon0)
    d_lat = np.radians(lat1 - lat0)
    lat0 = np.radians(lat0)
    lat1 = np.radians(lat1)

    a = hav(d_lat) + np.cos(lat0) * np.cos(lat1) * hav(d_lon)
    c = 2 * np.arcsin(np.sqrt(a))

    return radius * c


# ToDo: check whether this is really needed...
def approximate_connection_vector(lon0, lat0, lon1, lat1, len_lat_km=111.20):
    """Returns approximate connection vector between two points

    Note
    ----
    Careful: Only approximate, suited for small distances ( < 100km)

    Parameters
    ----------
    lon0 : float
        longitude of first point in decimal degrees
    lat0 : float
        latitude of first point in decimal degrees
    lon1 : float
        longitude of second point in decimal degrees
    lat1 : float
        latitude of second point in decimal degrees
    len_lat_km : float
        approx. length of 1 degree latitude at the equator in km.

    Returns
    -------
    numpy.ndarray
        2-element array containing (dx, dy) components of connection vector
    """

    lat = (lat0 + lat1) / 2 * 0.01745
    dx = len_lat_km * np.cos(lat) * (lon1 - lon0)
    dy = len_lat_km * (lat1 - lat0)
    return np.array((dx, dy))


def shifted_color_map(vmin, vmax, cmap=None):
    """Shift center of a diverging colormap to value 0

    Note
    ----

    This method was found `here <http://stackoverflow.com/questions/
    7404116/defining-the-midpoint-of-a-colormap-in-matplotlib>`__
    (last access: 17/01/2017). Thanks to `Paul H <http://stackoverflow.com/
    users/1552748/paul-h>`_ who provided it.

    Function to offset the "center" of a colormap. Useful for
    data with a negative min and positive max and if you want the
    middle of the colormap's dynamic range to be at zero level

    Parameters
    ----------
    vmin : float
        lower end of data value range
    vmax : float
        upper end of data value range
    cmap
        matplotlib colormap to be shifted (if None, use "seismic")

    Returns
    -------
    matplotlib.colors.LinearSegmentedColormap
        shifted colormap

    """
    import matplotlib.cm as colormaps
    import matplotlib.colors as colors
    if cmap is None:
        cmap = colormaps.seismic

    midpoint = 1 - np.abs(vmax) / (np.abs(vmax) + np.abs(vmin))

    cdict = {
        'red': [],
        'green': [],
        'blue': [],
        'alpha': []
    }

    # regular index to compute the colors
    reg_index = np.linspace(0, 1, 257)

    # shifted index to match the data
    shift_index = np.hstack([
        np.linspace(0.0, midpoint, 128, endpoint=False),
        np.linspace(midpoint, 1.0, 129, endpoint=True)
    ])

    for ri, si in zip(reg_index, shift_index):
        r, g, b, a = cmap(ri)

        cdict['red'].append((si, r, r))
        cdict['green'].append((si, g, g))
        cdict['blue'].append((si, b, b))
        cdict['alpha'].append((si, a, a))

    return colors.LinearSegmentedColormap('shiftedcmap', cdict)



# Geonum is a Python library for geographical calculations in 3D
# Copyright (C) 2017 Jonas Gliss (jonasgliss@gmail.com)
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License a
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
Module for atmospheric calculations relevant for geonum
"""
import pytest
import numpy as np
import numpy.testing as npt
from geonum import atmosphere as atm

def test_p0():
#: Pressure at sea level [Pa]
    assert atm.p0 == 101325.0

def test_NA():
    #: Avogadro constant [mol-1]
    assert atm.NA == 6.022140857e+23

def test_T0_STD():
    #: Standard sea level temperature [K]
    assert atm.T0_STD == 288.15

def test_M_AIR_AVG():
    #: Average molar mass of air, Unit: [g mol-1]
    assert atm.M_AIR_AVG == 28.9645

def test_R_STD():
    #: Gas constant (U.S. standard atmosphere) [Nm mol-1 K-1]
    assert atm.R_STD == 8.31432

def test_L_STD_ATM():
    #: Atmospheric lapse rate (Standard atmosphere) [K m-1]
    assert atm.L_STD_ATM == -6.5e-3

def test_L_DRY_AIR():
    #: Atmospheric lapse rate (dry atmosphere) [K m-1]
    assert atm.L_DRY_AIR == -9.8e-3

@pytest.mark.parametrize('phi,should_be',[
    (0, 9.780356),
    (0.79, 9.806398), #45deg
    (1.57, 9.83208), #90deg
    ])
def test__g0(phi,should_be):
    val = atm._g0(phi)
    npt.assert_allclose(val, should_be, rtol=1e-7)

@pytest.mark.parametrize('phi,alt,should_be',[
    (0, 0, 9.780356),
    (0.79, 0, 9.806398), #45deg
    (1.57, 0, 9.83208), #90deg
    (0, 1000, 9.471702),
    (0.79, 1000, 9.497973), #45deg
    (1.57, 1000, 9.523879), #90deg
    ])
def test__g_acc(phi,alt,should_be):
    val = atm._g_acc(phi,alt)
    npt.assert_allclose(val, should_be, rtol=1e-7)

@pytest.mark.parametrize('lat,alt,should_be',[
    (0, 0, 9.780356),
    (45, 0, 9.80616), #45deg
    (90, 0, 9.83208), #90deg
    (0, 1000, 9.471702),
    (45, 1000, 9.497733), #45deg
    (90, 1000, 9.523879), #90deg

    (np.array([45,90]),np.array([0,1000]),
     np.array([[9.80616, 9.83207964],
               [9.49773256, 9.5238791]]))
    ])
def test_g(lat,alt,should_be):
    val = atm.g(lat,alt)
    npt.assert_allclose(val, should_be, rtol=1e-7)

@pytest.mark.parametrize('lat,should_be',[
    (0, 9.780356),
    (45, 9.80616), #45deg
    (90, 9.83208), #90deg
    ])
def test_g0(lat,should_be):
    val = atm.g0(lat)
    npt.assert_allclose(val, should_be, rtol=1e-7)

@pytest.mark.parametrize('alt,ref_temp,ref_alt,lapse_rate,should_be',[
    (0,atm.T0_STD,0,atm.L_STD_ATM,atm.T0_STD),
    (0,atm.T0_STD,0,atm.L_STD_ATM,288.15),
    (0,atm.T0_STD,0,atm.L_DRY_AIR,288.15),
    (1000,atm.T0_STD,0,atm.L_DRY_AIR,278.35),
    (1000,atm.T0_STD,0,atm.L_STD_ATM,281.65),
    (1000,atm.T0_STD,100,atm.L_STD_ATM,282.3),
    (1000,atm.T0_STD,1000,atm.L_STD_ATM,288.15),
    ])
def test_temperature(alt,ref_temp,ref_alt,lapse_rate,should_be):
    val = atm.temperature(alt,ref_temp,ref_alt,lapse_rate)
    npt.assert_allclose(val, should_be, rtol=1e-7)

@pytest.mark.parametrize('mol_mass, lapse_rate, lat, should_be',[
    (atm.M_AIR_AVG, atm.L_STD_ATM, 45, -5.255632),
    (atm.M_AIR_AVG, atm.L_STD_ATM, 0, -5.241802),
    (atm.M_AIR_AVG, atm.L_STD_ATM, 90, -5.269523),
    (atm.M_AIR_AVG, atm.L_DRY_AIR, 45, -3.485878),
    ])
def test_beta_exp(mol_mass, lapse_rate, lat, should_be):
    val = atm.beta_exp(mol_mass, lapse_rate,lat)
    npt.assert_allclose(val, should_be, rtol=1e-7)

@pytest.mark.parametrize('alt,ref_p,ref_temp,ref_alt,lapse_rate,mol_mass,lat,should_be', [
    (0.0, atm.p0, atm.T0_STD, 0.0,atm.L_STD_ATM,atm.M_AIR_AVG,45.0,101325),
    (100.0, atm.p0, atm.T0_STD, 0.0,atm.L_STD_ATM,atm.M_AIR_AVG,45.0,100129.493856),
    (1000.0, atm.p0, atm.T0_STD, 0.0,atm.L_STD_ATM,atm.M_AIR_AVG,45.0,89875.07181),

    (0.0, 42, atm.T0_STD, 0.0,atm.L_STD_ATM,atm.M_AIR_AVG,45.0,42),
    (0.0, 42, 273.15, 0.0,atm.L_STD_ATM,atm.M_AIR_AVG,45.0,42),
    (100.0, 42, 273.15, 0.0,atm.L_STD_ATM,atm.M_AIR_AVG,45.0,41.477378),

    (0.0, atm.p0, 273.15, 0.0,atm.L_STD_ATM,atm.M_AIR_AVG,45.0,101325),
    (100.0, atm.p0, 273.15, 0.0,atm.L_STD_ATM,atm.M_AIR_AVG,45.0,100064.17512),

    (0.0, atm.p0, 273.15, 100.0,atm.L_STD_ATM,atm.M_AIR_AVG,45.0,102598.657973),

    (0.0, atm.p0, 273.15, 0,atm.L_DRY_AIR,atm.M_AIR_AVG,45.0,101325),
    (100.0, atm.p0, atm.T0_STD, 0.0,atm.L_DRY_AIR,atm.M_AIR_AVG,45.0,100128.81154),

    (0.0, atm.p0, atm.T0_STD, 0.0,atm.L_STD_ATM,20,45.0,101325),
    (100.0, atm.p0, atm.T0_STD, 0.0,atm.L_STD_ATM,20,45.0,100497.98743),

    (0.0, atm.p0, atm.T0_STD, 0.0,atm.L_STD_ATM,atm.M_AIR_AVG,0,101325),
    (100.0, atm.p0, atm.T0_STD, 0.0,atm.L_STD_ATM,atm.M_AIR_AVG,0,100132.621128),
    (0.0, atm.p0, atm.T0_STD, 0.0,atm.L_STD_ATM,atm.M_AIR_AVG,90,101325),
    (100.0, atm.p0, atm.T0_STD, 0.0,atm.L_STD_ATM,atm.M_AIR_AVG,90,100126.352659),
    ])
def test_pressure(alt,ref_p,ref_temp,ref_alt,lapse_rate,mol_mass,lat,should_be):
    val = atm.pressure(alt,ref_p,ref_temp,ref_alt,lapse_rate,mol_mass,lat)
    npt.assert_allclose(val, should_be, rtol=1e-7)


@pytest.mark.parametrize('alt,ref_p,ref_temp,ref_alt,lapse_rate,mol_mass,lat,should_be', [
    (0.0, atm.p0, atm.T0_STD, 0.0,atm.L_STD_ATM,atm.M_AIR_AVG,45.0,101325/100),
    (100.0, atm.p0, atm.T0_STD, 0.0,atm.L_STD_ATM,atm.M_AIR_AVG,45.0,100129.493856/100),
    (1000.0, atm.p0, atm.T0_STD, 0.0,atm.L_STD_ATM,atm.M_AIR_AVG,45.0,89875.07181/100),
    ])
def test_pressure_hPa(alt,ref_p,ref_temp,ref_alt,lapse_rate,mol_mass,lat,should_be):
    val = atm.pressure_hPa(alt,ref_p,ref_temp,ref_alt,lapse_rate,mol_mass,lat)
    npt.assert_allclose(val, should_be, rtol=1e-7)

@pytest.mark.parametrize('p, ref_p, ref_temp, ref_alt, lapse_rate, mol_mass, lat, should_be', [
    (atm.p0, atm.p0, atm.T0_STD,0,atm.L_STD_ATM,atm.M_AIR_AVG,45,0),
    (100000, atm.p0, atm.T0_STD,0,atm.L_STD_ATM,atm.M_AIR_AVG,45,110.889658),
    (50000, atm.p0, atm.T0_STD,0,atm.L_STD_ATM,atm.M_AIR_AVG,45,5574.679741),

    (atm.p0, atm.p0, 42,0,42,42,42,0),
    (50000, 50000, atm.T0_STD,0,atm.L_STD_ATM,atm.M_AIR_AVG,45,0),
    (50000, 80000, atm.T0_STD,0,atm.L_STD_ATM,atm.M_AIR_AVG,45,3792.338559),

    (atm.p0, atm.p0, 273.15,0,atm.L_STD_ATM,atm.M_AIR_AVG,45,0),
    (50000, atm.p0, 273.15,0,atm.L_STD_ATM,atm.M_AIR_AVG,45,5284.482982),
    (50000, atm.p0, 300,0,atm.L_STD_ATM,atm.M_AIR_AVG,45,5803.935181),

    (atm.p0, atm.p0, atm.T0_STD,100,atm.L_STD_ATM,atm.M_AIR_AVG,45,100),

    (atm.p0, atm.p0, atm.T0_STD,0,atm.L_DRY_AIR,atm.M_AIR_AVG,45,0),
    (50000, atm.p0, atm.T0_STD,0,atm.L_DRY_AIR,atm.M_AIR_AVG,45,5392.870582),

    (atm.p0, atm.p0, atm.T0_STD,0,atm.L_STD_ATM,20,45,0),
    (50000, atm.p0, atm.T0_STD,0,atm.L_STD_ATM,20,45,7840.324557),

    (atm.p0, atm.p0, atm.T0_STD,0,atm.L_STD_ATM,atm.M_AIR_AVG,45,0),
    (atm.p0, atm.p0, atm.T0_STD,0,atm.L_STD_ATM,atm.M_AIR_AVG,0,0),
    (atm.p0, atm.p0, atm.T0_STD,0,atm.L_STD_ATM,atm.M_AIR_AVG,90,0),

    (50000, atm.p0, atm.T0_STD,0,atm.L_STD_ATM,atm.M_AIR_AVG,0,5588.419043),
    (50000, atm.p0, atm.T0_STD,0,atm.L_STD_ATM,atm.M_AIR_AVG,90,5560.946563),

    ])
def test_pressure2altitude(p, ref_p, ref_temp, ref_alt, lapse_rate, mol_mass, lat, should_be):
    val = atm.pressure2altitude(p, ref_p, ref_temp, ref_alt, lapse_rate,
                                mol_mass, lat)
    npt.assert_allclose(val, should_be, rtol=1e-7)

@pytest.mark.parametrize('alt,temp,ref_p,ref_temp,ref_alt,lapse_rate,mol_mass,lat,should_be', [

    (0,None,atm.p0,atm.T0_STD,0,atm.L_STD_ATM,atm.M_AIR_AVG,45,1225.0033852145957),
    (100,None,atm.p0,atm.T0_STD,0,atm.L_STD_ATM,atm.M_AIR_AVG,45,1213.286799),
    (1000,None,atm.p0,atm.T0_STD,0,atm.L_STD_ATM,atm.M_AIR_AVG,45,1111.65185),
    (10000,None,atm.p0,atm.T0_STD,0,atm.L_STD_ATM,atm.M_AIR_AVG,45,412.733471),

    (0,273.15,atm.p0,atm.T0_STD,0,atm.L_STD_ATM,atm.M_AIR_AVG,45,1292.274301),
    (0,300,atm.p0,atm.T0_STD,0,atm.L_STD_ATM,atm.M_AIR_AVG,45,1176.615751),

    (0,None,50000,atm.T0_STD,0,atm.L_STD_ATM,atm.M_AIR_AVG,45,604.492171),
    (0,None,50000,251.9145816835,5574.679741,atm.L_STD_ATM,atm.M_AIR_AVG,45,1225.0033852145957),

    ])
def test_air_density(alt,temp,ref_p,ref_temp,ref_alt,lapse_rate,mol_mass,lat,should_be):
    val = atm.air_density(alt,temp,ref_p,ref_temp,ref_alt,lapse_rate,mol_mass,lat)
    npt.assert_allclose(val, should_be, rtol=1e-7)

@pytest.mark.parametrize('alt,temp,ref_p,ref_temp,ref_alt,lapse_rate,mol_mass,lat,should_be', [

    (0,None,atm.p0,atm.T0_STD,0,atm.L_STD_ATM,atm.M_AIR_AVG,45,2.5469602223632817e+25),
    (10000,None,atm.p0,atm.T0_STD,0,atm.L_STD_ATM,atm.M_AIR_AVG,45,8.581329e+24),
    ])
def test_air_number_density(alt,temp,ref_p,ref_temp,ref_alt,lapse_rate,mol_mass,lat,should_be):
    val = atm.air_number_density(alt,temp,ref_p,ref_temp,ref_alt,lapse_rate,mol_mass,lat)
    npt.assert_allclose(val, should_be, rtol=1e-7)

@pytest.mark.parametrize('lbda_mu,should_be', [
    (0.3, 1.000291554210872),
    (0.2, 1.000324)
    ])
def test_refr_idx_300ppm_co2(lbda_mu,should_be):
    val = atm.refr_idx_300ppm_co2(lbda_mu)

    npt.assert_allclose(val, should_be, rtol=1e-7)

def test_refr_idx():
    val = atm.refr_idx()
    should_be=1.000291554210872
    npt.assert_allclose(val, should_be, rtol=1e-7)

def test__F_N2():
    val = atm._F_N2()
    should_be=1.037522
    npt.assert_allclose(val, should_be, rtol=1e-5)

def test__F_O2():
    val = atm._F_O2()
    should_be=1.129265
    npt.assert_allclose(val, should_be, rtol=1e-5)

def test_F_AIR():
    val = atm.F_AIR()
    should_be=1.056395
    npt.assert_allclose(val, should_be, rtol=1e-5)

@pytest.mark.parametrize('lbda_mu,co2_ppm,should_be', [
    (0.3, 400,5.652217e-26),
    (0.5, 400,6.661139e-27),
    (0.9, 400,6.133698e-28),

    (0.5, 500,6.661859e-27),

    (0.5, 100,6.658979e-27),
    ])
def test_sigma_rayleigh(lbda_mu,co2_ppm,should_be):
    val = atm.sigma_rayleigh(lbda_mu,co2_ppm)
    npt.assert_allclose(val, should_be, rtol=1e-5)

@pytest.mark.parametrize('alt,lbda_mu,co2_ppm,should_be', [
    (0,0.3,400,1.439597e-06),
    (1000,0.3,400,1.306389e-06),
    (10000,0.3,400,4.850354e-07),

    (0,0.5,400,1.696566e-07),
    (0,0.9,400,1.562228e-08),

    (0,0.3,400,1.439597e-06),
    ])
def test_rayleigh_vol_sc_coeff(alt,lbda_mu,co2_ppm,should_be):
    val = atm.rayleigh_vol_sc_coeff(alt,lbda_mu,co2_ppm)
    npt.assert_allclose(val, should_be, rtol=1e-5)


import glob
import os

import srtm


def get_local_srtm_files():
    """
    Get list of locally stored SRTM data files

    Returns
    -------
    list
        list of file paths
    """
    fh = srtm.utils.FileHandler()
    return glob.glob(f'{fh.local_cache_dir}/*.hgt')


def delete_local_srtm_files():
    """Delete all locally stored SRTM files"""
    for file in get_local_srtm_files():
        print(f'Deleting SRTM data file at {file}')
        os.remove(file)



# -*- coding: utf-8 -*-
import matplotlib
matplotlib.use('Agg')
import pytest
from contextlib import contextmanager
from geonum import NETCDF_AVAILABLE


# custom skipif marker that is used below for test functions that
# require SRTM.py to be installed
def srtm_works():
    """SRTM access is broken, see
    https://github.com/tkrajina/srtm.py/issues/51
    """
    import srtm
    try:
        srtm.get_data().get_file(50,8)
        return True
    except Exception as e:
        print('failed to access SRTM data...')
        if not str(e).startswith('Cannot retrieve https'):
            raise
        return False


skip_srtm = pytest.mark.skipif(not srtm_works(),
                               reason=('Skipping SRTM database tests since '
                                       'SRTM data cannot be accessed.'))

skip_netcdf4 = pytest.mark.skipif(NETCDF_AVAILABLE==False,
                                  reason='Skipping tests requiring NetCDF4 library.')


@contextmanager
def does_not_raise_exception():
    yield



# Geonum is a Python library for geographical calculations in 3D
# Copyright (C) 2017 Jonas Gliss (jonasgliss@gmail.com)
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License a
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import warnings

from LatLon23 import GeoVector, LatLon
from numpy import (radians, cos, sin, degrees, sqrt, tan, isnan, arctan2)

from geonum.geopoint import GeoPoint


class GeoVector3D(GeoVector):
    """A 3-dimensional geo vector object
    
    3D vector representation for geodesic connections and arithmetics of
    locations (:class:`geonum.GeoPoint`).

    Note
    ----
    To create an instnace of :class:`GeoVector3D` use one of the following to
    input combinations:

        1. dx, dy, dz
        #. dz, azimuth, dist_hor
        #. azimuth, dist_hor, elevation

    Multiple input combinations will be processed in the preferred order
    as given in the list.


    Note
    ----
    Horitontal displacement components dx (longitude) and dy (latitude) are
    in units of km, while vertical displacement component dz (altitude) is
    in units of m.

    Attributes
    ----------
    dx : float
        longitudinal component in units of km
    dy : float
        latitudinal component in units of km
    dz : float
        altitude component in units of m
    name : str
        name of vector


    Parameters
    ----------
    dx : float, optional
        longitudinal component in units of km
    dy : float, optional
        latitudinal component in units of km
    dz : float, optional
        altitude component in units of m
    azimuth : float, optional
        azimuth orientation angle in units of decimal degrees, relative to
        North direction.
    dist_hor : float, optional
        horizontal displacement length in input `azimuth` direction.
    elevation : float, optional
        elevation angle in decimal degrees (0 corresponds to horizon,
        90 to zenith)
    anchor : GeoPoint, optional
        anchor point of this vector.
    name : str, optional
        name of vector, defaults to "undefined".

    Example
    -------

    >>> from geonum import GeoPoint, GeoVector3D
    >>> p = GeoPoint(latitude=10.0, longitude=15.0, name="random_point")
    >>> v = GeoVector3D(dx=15, dy=100, dz=300) #dx, dy in km, dz in m
    >>> new_point = p + v # GeoPoint object
    >>> print(new_point)
    GeoPoint undefined
    Lat: 10.904041412793307,  Lon: 15.137202747069097, Alt: 300 m

    """

    def __init__(self, dx=None, dy=None, dz=None, azimuth=None, dist_hor=None,
                 elevation=None, anchor=None, name=None):
        if name is None:
            name = 'undefined'

        self.name = name

        self.dx = None
        self.dy = None
        self.dz = None

        self._eval_input(dx,dy,dz,azimuth,dist_hor,elevation)

        # dictionary with private attributes
        self._priv_attr= {"anchor" : None}

        # call setter for private attribute anchor (this ensures that input
        # attr anchor is of right type
        if anchor is not None:
            try:
                self.set_anchor(anchor)
            except TypeError as e:
                print(f'Failed to set anchor. Reason: {e}')


    def _eval_input(self,dx,dy,dz,azimuth,dist_hor,elevation):
        """
        Retrieve dx, dy, dz based on input

        See parameter description regarding allowed input combinations. Note that for the horizontal
        component, inputs of `dx` and `dy` are preferred over `azimuth` and `elevation` (in case
        both are wrongly provided).

        Parameters
        ----------
        dx : float, optional
            delta x in units of longitude degrees East.
            To be combined with `dy`. If None, it is expected that
            azimuth and dist_hor are provided.
        dy : float, optional
            delta y in units of latitude North, to be combined with `dx`.
            If None, it is expected that azimuth and dist_hor are provided.
        dz : float, optional
            delta z in units of m. If None or NaN, it will either be inferred
            from input `elevation` and `dx` & `dy` or if that's not
            possible (`elevation` is None), then it will be set to 0 m.
        azimuth : float, optional
            horizontal direction of vector in units of radians or degrees
            (relative to North, or 12 o'clock direction)
        dist_hor : float, optional
            magnitude of horizontal component of vector (dx, dy).
        elevation : float, optional
            vertical elevation angle in units of degrees. Allowed input range are values between
            (and excluding) -90 and +90 (zenith).
        """
        if dx is None or dy is None:
            if azimuth is not None and dist_hor is not None:
                theta_rad = radians(self._angle_or_heading(azimuth))
                dx = dist_hor * cos(theta_rad)
                dy = dist_hor * sin(theta_rad)
            else: # If only dx and dy are given
                raise ValueError('Please specify either: 1. azimuth and dist_hor, or 2. dx and dy')

        self.dx = dx
        self.dy = dy
        # Check input for altitude component dz
        if dz is None or isnan(dz): #invalid for dz directly
            if elevation is not None and -90 < elevation < 90: #check if instead elevation is valid, then set dz
                dz = tan(radians(elevation))*sqrt(self.dx**2+self.dy**2)*1000
            else: #both dz input and elevation are invalid, set dz=0
                dz = 0.0
        self.dz = dz

    @property
    def dz_km(self) -> float:
        """:attr:`dz` converted to units of km

        E.g. used for internal arithmetics.
        """
        return self.dz/1000

    @property
    def azimuth(self) -> float:
        """Horizontal orientation angle relative to North direction"""
        return degrees(arctan2(self.dx, self.dy))

    @property
    def elevation(self) -> float:
        """Elevation angle in decimal degrees relative to horizon"""
        return (90-self.polar_angle)
    
    @property
    def polar_angle(self) -> float:
        """Polar angle in decimal degrees relative to zenith"""
        return degrees(arctan2(self.dist_hor, self.dz_km))
    
    @property    
    def magnitude(self) -> float:
        """Magnitude of vector (length) in units of km"""
        return sqrt(self.dist_hor**2 + self.dz_km**2)

    @property
    def norm(self) -> float:
        """Norm of vector, wrapper for :meth:`magnitude`"""
        return self.magnitude
        
    @property
    def dist_hor(self) -> float:
        """Horizontal distance spanned by this vector"""
        return sqrt(self.dx**2 + self.dy**2)
    
    @property
    def anchor(self) -> GeoPoint:
        """Anchor point of vector"""
        return self._priv_attr["anchor"]
        
    @anchor.setter
    def anchor(self, value):
        self.set_anchor(value)
    
    def set_anchor(self, geo_point):
        """Set anchor of this vector

        Parameters
        ----------
        geo_point : GeoPoint
            anchor location.

        Raises
        ------
        TypeError
            if input point is not instance of :class:`GeoPoint`.
        """
        if not isinstance(geo_point, GeoPoint):
            if isinstance(geo_point, LatLon):
                geo_point = GeoPoint.from_LatLon(geo_point)
            else:
                raise TypeError("Could not set anchor: Invalid input type")
        self._priv_attr["anchor"] = geo_point
           
    def intersect_hor(self, other) -> 'GeoVector3D':
        """Find horizontal intersection of this vector with another vector

        Note
        ----
        Only works if anchor point (:attr:`anchor`) is set in both vectors.

        Parameters
        ----------
        other : GeoVector3D
            Other vector for which the intersection is to be determined.

        Raises
        ------
        AttributeError
            if :attr:`anchor` is not set in this vector or input vector

        Returns
        -------
        GeoVector3D
            New vector pointing in direction of this vector but with correct
            horizontal magnitude to the intersection with the other vector.

        """
        if not all([isinstance(x,GeoPoint) for x in [self.anchor, other.anchor]]):
            raise AttributeError(
                "Intersection can not be determined, anchor "
                "of one of the vectors not set..")
        v = other.anchor - self.anchor
        v.dz = 0.0
        other_az = radians((other.azimuth + 360) % 360)
        self_az = radians((self.azimuth + 360) % 360)
        dy = (v.dx - tan(other_az) * v.dy) / (tan(self_az) - tan(other_az))
        dx = tan(self_az) * dy
        return GeoVector3D(dx, dy, dz=0.0)

    def plot(self, map, add_anchor=False, **kwargs): # pragma: no cover
        """Plot this vector into existing basemap"""
        warnings.warn(DeprecationWarning(
            'See https://github.com/jgliss/geonum/issues/4'))
        map.draw_geo_vector_3d(self, **kwargs)
        if add_anchor:
            self.anchor.plot(map, add_name=True, dz_text=self.dz*.1)


    # ToDo: check if __call__ is really needed, a little confusing...
    def __call__(self):
        """Call function

        Returns
        -------
        float
            azimuth angle in units of decimal degrees wrt to N direction
        float
            horizontal length in km
        float
            vertical length in m
        """
        return (self.azimuth, self.dist_hor, self.dz)
        
    def __neg__(self):
        """Returns negative of this vector"""
        return GeoVector3D(-self.dx, -self.dy, -self.dz)

    def __truediv__(self, other):
        if isinstance(other, GeoVector3D):
            dx = self.dx / other.dx
            dy = self.dy / other.dy
            dz = self.dz / other.dz
        else:
            dx = self.dx / other
            dy = self.dy / other
            dz = self.dz / other
        return GeoVector3D(dx,dy,dz)


    def __add__(self, other):
        """Add another geo vector"""
        return GeoVector3D(self.dx + other.dx, self.dy + other.dy,
                           self.dz + other.dz)

    def __str__(self) -> str:
        """String representation"""
        return ('GeoVector3D {}\n'
                'Azimuth: {:.2f}°, Elevation: {:.4f}°, Magnitude: {:.2f} km '
                '(hor: {:.2f} km)'
                .format(self.name, self.azimuth, self.elevation, 
                        self.magnitude, self.dist_hor))
            
    
    def __repr__(self) -> str:
        return ('Az %s, Elev %s, Mag. %s' 
                %(self.azimuth, self.elevation, self.magnitude))
    
    def type(self) -> str:
        """Returns object type"""
        return 'GeoVector3D'



# -*- coding: utf-8 -*-
"""Test environment for base.py module."""
import numpy.testing as npt
import pytest

from geonum import GeoPoint, GeoVector3D


def test_GeoPoint():
    """Test basic arithmetic operations on GeoPoints."""
    p = GeoPoint(latitude=0, longitude=0,
                 auto_topo_access=False)
    assert p.latitude == 0
    assert p.longitude == 0


def test_GeoVector3D():
    """Test basic arithmetic operations on GeoVectors."""
    v = GeoVector3D(dx=1, dy=1, dz=100)

    npt.assert_array_equal([v.dx, v.dy, v.dz], [1, 1, 100])
    npt.assert_array_almost_equal([v.elevation,
                                   v.magnitude,
                                   v.azimuth,
                                   v.dist_hor],
                                  [4.044691, 1.417745, 45., 1.414214])


def test_diffvector():
    p1 = GeoPoint(latitude=37.751005, longitude=14.993435,
                  altitude=3264.0,
                  auto_topo_access=False)
    p2 = GeoPoint(37.765755, 15.016696, altitude=2820.0,
                  auto_topo_access=False)
    connection_vector = p2 - p1

    assert connection_vector.anchor is p1

    actual = [connection_vector.azimuth,
              connection_vector.elevation,
              connection_vector.magnitude]

    npt.assert_allclose(actual=actual,
                        desired=[51.378677249653983,
                                 -9.6064174085658465,
                                 2.6606074796318557], rtol=1e-7)


if __name__ == "__main__":
    import sys

    pytest.main(sys.argv)



import numpy as np
import pytest
import geonum.geovector3d as mod

def test_init_NO_INPUT():
    with pytest.raises(ValueError):
        mod.GeoVector3D()

def test_init_DOUBLE_INPUT():
    v = mod.GeoVector3D(dx=1, dy=1, azimuth=223, dist_hor=1.2456)
    assert v.dx == 1
    assert v.dy == 1
@pytest.mark.parametrize('kwargs,mag,azim,elev', [

    (dict(dx=2, dy=0, dz=0),2,90,0),
    (dict(dx=2, dy=0, elevation=0),2,90,0),
    (dict(dx=2, dy=0, elevation=90),2,90,0),
    (dict(dx=2, dy=0, elevation=45),2**(1.5),90,45),
    (dict(dx=1, dy=-1, dz=0),2**(0.5),135,0),
    (dict(azimuth=140,dist_hor=111,dz=0),111,140,0),
    (dict(azimuth=140,dist_hor=1,dz=1000),2**0.5,140,45),
])
def test_init(kwargs,mag,azim,elev):
    v = mod.GeoVector3D(**kwargs)
    _mag = v.magnitude
    np.testing.assert_allclose(v.magnitude, mag, rtol=1e-5)
    np.testing.assert_allclose(v.azimuth, azim, rtol=1e-5)
    np.testing.assert_allclose(v.elevation, elev, rtol=1e-5)
def test_truediv_CASE1():
    v1 = mod.GeoVector3D(dx=1,dy=1,dz=1)
    v2 = mod.GeoVector3D(dx=1, dy=1, dz=1)

    res = v1.__truediv__(v2)
    assert isinstance(res, mod.GeoVector3D)
    assert res.magnitude == v1.magnitude

def test_truediv_CASE2():
    v1 = mod.GeoVector3D(dx=1,dy=1,dz=1)

    res = v1.__truediv__(1)
    assert isinstance(res, mod.GeoVector3D)
    assert res.magnitude == v1.magnitude


import pytest
import os, shutil
import numpy as np
import numpy.testing as npt
from geonum import LOCAL_TOPO_DIR, TOPO_INFO_FILE
from geonum import helpers as h

@pytest.mark.parametrize('num,result', [
    (10, 1), (1000, 3), (1.2,0), (0.023, -2)
])
def test_order_of_magnitude(num,result):
    assert h.order_of_magnitude(num) == result

def test_all_topodata_search_dirs():
    all_dirs = h.all_topodata_search_dirs()
    assert isinstance(all_dirs, list)
    assert LOCAL_TOPO_DIR in all_dirs
    for dirloc in all_dirs:
        assert os.path.exists(dirloc)

def test_check_and_add_topodir():
    tmp_file = os.path.join(LOCAL_TOPO_DIR, 'TMP.file')
    topofile_tmp = shutil.copyfile(TOPO_INFO_FILE, tmp_file)
    all_dirs = h.all_topodata_search_dirs()
    num_init = len(all_dirs)
    h.check_and_add_topodir(LOCAL_TOPO_DIR)
    assert len(h.all_topodata_search_dirs()) == num_init # it is already in there and should thus not be added

    homedir = os.path.expanduser('~')
    if not homedir in all_dirs:
        h.check_and_add_topodir(homedir)
        assert len(h.all_topodata_search_dirs()) == num_init + 1

    invaliddir = 'bla/blub.xuspf/nla'
    if os.path.exists(invaliddir):
        print('What a surprise indeed... path {invaliddir} exists :-O')
    else:
        with pytest.raises(ValueError):
            h.check_and_add_topodir(invaliddir)

    # go back to initial state
    topofile_tmp = shutil.copyfile(tmp_file, TOPO_INFO_FILE)
    os.remove(tmp_file)

@pytest.mark.parametrize('val,expectation', [
    (0, True), (0.1, True), ('1', False),
    ({}, False), ([1], False), ((1,2,), False)
    ])
def test_isnum(val,expectation):
    assert h.isnum(val) == expectation


@pytest.mark.parametrize('lon0,lat0,lon1,lat1,radius,value', [
    (-77.037852,38.898556,-77.043934,38.897147,None,0.549156),
    (-90,0,90,0,None,6371 * np.pi), # half around the equator
    (-90,90,90,-90,None,6371 * np.pi), # pole to pole
    (-90,0,90,0,500,500 * np.pi), # half around the equator
    ])
def test_haversine_formula(lon0,lat0,lon1,lat1,radius,value):
    val = h.haversine_formula(lon0, lat0, lon1, lat1, radius)
    npt.assert_allclose(val, value, rtol=1e-5)


if __name__ == "__main__":
    import sys
    pytest.main(sys.argv)


